.586
.model flat,stdcall 
option casemap:none 
include \masm32\include\windows.inc 
include \masm32\include\kernel32.inc 
include \masm32\include\user32.inc 
include \masm32\m32lib\masm32.inc 
includelib \masm32\lib\kernel32.lib 
includelib \masm32\lib\user32.lib
includelib \masm32\m32lib\masm32.lib

.data 
align 16
  brt    dd  32 dup (0)  ; really use 512+1 bytes
  brm    dd  32 dup (0)  ;
  bsp    db 512 dup (0)  ; 1024 bits  512 bytes  4 bytes/byte 2,2,2,2 : 8
  bdd    dd  32 dup (0)  ; 1024 bits  128 bytes 32 dwords
  jbd    dd  31
  jbr    dd  16   ; (jbd / 2) + 1
  jbs    dd 127   ; (jbd+1 * 4) - 1  
  ksp    dd   0
  k      dd   0  
  j      dd   0
  exact  dd   0

  szCaption  db "(Unsigned) Integer Square Root with Remainder",0 
  szFmtStr   db "%lu iterations",0
  szBuffer   db 32 dup (0)

; 
;   
;   Large Integer Square Root
;
;
; This program takes a Large Unsigned Integer 
; and calculates the Integer Square Root and Remainder.
;
; Does 1 Million square root calculations in 157 seconds of CPU time.
; 1.2 Ghz AMD Athlon 32 bit CPU using Windows XP.
;
;
; The routine can scale to various size integers 
; by adjusting the size of the main variables 
; and the variables that record their sizes.
;
;
; bdd the Number itself       x bytes or  x/4 dwords
; bsp the Number spread out   2 bits at a time    
;                           bsp 76 54 32 10  in 4 bytes
;                           bdd 76543210     in 1 byte
; brt the Root
; brm the Remainder
;
; jbd number of dwords 0 based bdd   ie 32 dwords  0..31  so jbd == 31
; jbr number of dwords 0 based br?   both brt and brm  (jbd div 2) + 1
;
; jbs number of dwords 0 based bsp   is also the number of bytes of bdd
;
; ksp number of bytes used by bsp
; k   loop counter to use each byte of bsp
;
; j   counter
; exact  flag  1 if an exact Square Root ie No remainder,  0 otherwise
;
; It can also detect when the root is exact  ie Square Root 10 for 100
;
;
;   Fairly close Delphi/Pascal code using a 16 bit / 2 byte number.
;   The sp filling routine is the only difference, 
;   The MASM code doesn't shift, pulls in place.
;
;   dd := 65535;                        (* 16 bit number     *)
;   rm := 0;
;   rt := 0;
;   for i := 0 to 7 do                  (* 0 to (bits / 2)-1 *)
;     sp[i] := (dd shr (i*2)) and 3;    (* 2 bits at a time  *)
; 
;   for i := 7 downto 0 do begin
;     rt :=  (rt shl 1);                (* rt * 2            *)
;     rm :=  (rm shl 2) + sp[i];        (* rm * 4  +  sp[i]  *)
; 
;     if (rt <  rm) then begin
;       rm := rm - rt - 1;
;       rt := rt + 2;
;     end;
;   end;
;   rt := rt shr 1;                     (* rt / 2  when done *)
;
;   exact := 0;
;   if rm = 0) and (rt*rt = dd) then
;     exact := 1;
;
;   writeln('Number    : ',dd);
;   writeln('Root      : ',rt);
;   writeln('Remainder : ',rm);
;
; 






.code 
start:

; 
; bdd  initialize with a fixed set of bytes
; 

  mov esi, 0
  .repeat
     mov eax, 3589713
     invoke nrandom, eax
     mov byte ptr [bdd+esi], al
     inc esi
  .until esi > 127    ; db 127 hi of dd


; 
; bdd  zero the low dword, set the high byte to Fx
; 

  mov bdd, 0
  or  byte ptr [bdd+127], 240

; 
; 
; 






; 
; Do a test of 1,000,000 iterations of Integer Square Root
; 

  .while (bdd < 1000000)

; 
; brt brm  zero them out
; 

  mov ecx, jbr     ; 31 max   Use 16  which covers 65 bytes
  xor eax, eax
  .repeat
     mov [brt+ecx*4], eax
     mov [brm+ecx*4], eax
  .untilcxz
  mov brt, eax
  mov brm, eax

; 
; bsp  zero it    Can be skipped because it will overwrite during fill
; 

 ;   mov ecx, jbs  ; 0 to 127 dd,  0 to 511 db  
 ;   xor eax, eax
 ;   .repeat
 ;      mov dword ptr [bsp+ecx*4], eax
 ;   .untilcxz
 ;   mov dword ptr bsp, eax

; 
; bsp  fill it with  bdd  two bits at a time
; 

  xor esi, esi  ; 0 to 127 max
  xor edi, edi  ; 0 to 511 max
  xor eax, eax
  .repeat
    mov dl, byte ptr [bdd+esi]

    mov al, dl
;    shr al, 0
    and al, 3
    mov [bsp+edi], al
    inc edi

    mov al, dl
    shr al, 2
    and al, 3
    mov [bsp+edi], al
    inc edi

    mov al, dl
    shr al, 4
    and al, 3
    mov [bsp+edi], al
    inc edi

    mov al, dl
    shr al, 6    ; could shl 2 then use ah instead
;    and al, 3
    mov [bsp+edi], al
    inc edi

    inc esi
  .until esi > jbs  ; for every 1 byte of bdd, 4 bytes of bsp
  dec edi

; 
; adjust  ksp to highest byte in bsp with value > 0
; 

  .if [bsp+edi] == 0
     dec edi
    .if [bsp+edi] == 0
       dec edi
       .if [bsp+edi] == 0
          dec edi
       .endif
    .endif
  .endif
  mov ksp, edi







; 
; for k = ksp downto 0
; 

  mov ebx, ksp
  mov k, ebx
  .while  k < 0ffffffffh   ; ? maybe ?   sdword ptr k >= 0 

; 
; brt * 2
; 

  mov edx, jbr
  dec edx
  mov ecx, jbr
  .repeat
     mov  eax, [brt+edx*4]
     shld [brt+ecx*4], eax, 1
     dec  edx
  .untilcxz

; 
; brm * 4   + bsp[k]
; 

  mov edx, jbr
  dec edx
  mov ecx, jbr
  .repeat
     mov  eax, [brm+edx*4]
     shld [brm+ecx*4], eax, 2
     dec  edx
  .untilcxz
  mov ecx, k
  mov al, [bsp+ecx]
  or  byte ptr brm, al   ; brm 7..2  bsp[k] 1..0
;  add byte ptr brm, al

; 
; compare  brt with brm    > = <   3, 2, 1
; 

  mov j, 2
  mov ecx, jbr
  inc ecx
  mov edx, jbr
  .repeat
    mov esi, [brt+edx*4]
    mov edi, [brm+edx*4]
    .if esi > edi
       mov j, 3
    .elseif esi < edi
       mov j, 1
    .endif
  .untilcxz (j != 2)

; 
; brm = brm - brt - 1
; 

  .if eax == 1
    mov esi, 0
    stc
    .repeat
      mov eax, [brt+esi*4]
      sbb [brm+esi*4], eax
      inc esi
    .until esi > jbr
  
; 
; brt = brt + 2
; 

    add brt, 2
  .endif
  
; 
; endfor
; 

    dec k
  .endw


; 
; brt / 2   Square root calc almost done just scale down brt (shr 1)
; 

  mov edx, 0
  mov ecx, 1
  .while edx <= jbr
     mov  eax, [brt+ecx*4]
     shrd [brt+edx*4], eax, 1
     inc  edx
     inc  ecx
  .endw


; 
; Is it an exact SQUARE ROOT ?
; 

  mov exact, 0
  xor edx, edx
  mov eax, brt
  mul brt
  .if (eax == bdd)
     xor edx, edx
     xor eax, eax
     add eax, brm+0
     adc edx, 0
     add eax, brm+4
     adc edx, 0
     add eax, brm+8
     adc edx, 0
     add eax, brm+12
     adc edx, 0
     .if (edx == 0) && (eax == 0)
        mov exact, 1
     .endif
  .endif

; 
; increment the iterations loop
; 

    inc bdd  ; low dword of bdd doing double duty as loop var
  .endw

; 
; 
; 


; 
; 
; 


  invoke wsprintf,addr szBuffer,addr szFmtStr, bdd
  invoke MessageBox, NULL, addr szBuffer, addr szCaption, MB_OK

  invoke ExitProcess, NULL 
end start